import {Component, ElementRef, ViewChild, OnInit, Input, OnDestroy, HostListener} from '@angular/core';

import {
  Label,
  MySprite, ShapeCircle, ShapeRect, ShapeRectNew, tweenChange,

} from './Unit';
import {res, resAudio} from './resources';

import {Subject} from 'rxjs';
import {debounceTime} from 'rxjs/operators';

import TWEEN from '@tweenjs/tween.js';




@Component({
  selector: 'app-play',
  templateUrl: './play.component.html',
  styleUrls: ['./play.component.scss']
})
export class PlayComponent implements OnInit, OnDestroy {

  @ViewChild('canvas') canvas: ElementRef;
  @ViewChild('wrap') wrap: ElementRef;

  // 数据
  data;

  ctx;

  canvasWidth = 1280; // canvas实际宽度
  canvasHeight = 720; // canvas实际高度

  canvasBaseW = 1280; // canvas 资源预设宽度
  canvasBaseH = 720;  // canvas 资源预设高度

  mx; // 点击x坐标
  my; // 点击y坐标


  // 资源
  rawImages = new Map(res);
  rawAudios = new Map(resAudio);

  images = new Map();

  animationId: any;
  winResizeEventStream = new Subject();

  audioObj = {};

  renderArr;
  mapScale = 1;

  canvasLeft;
  canvasTop;

  saveKey = 'test_008';

  pause;
  btnLeft;
  btnRight;
  pic1;
  pic2;

  canTouch = true;

  curPic;

  //单词画面数组
  word_arr_left = [] ;
  word_arr = [];

  word_bg_x ;

  word_height ;

  all_words_loaded = false;

  curAudio ;
  play = false;
  words_bg ;
  title_bg ;

  wordLoad :{[key:string]:{loaded : boolean;letters:{[key:string]:{loaded:boolean, letter:Label}}}} = {};

  setWordLoad(id: string, letterId:string, loaded:boolean, letter: Label):void{

    if(!this.wordLoad[id]){
      this.wordLoad[id] = {
        loaded:false,
        letters:{}
      }
    }
    let letters = this.wordLoad[id].letters;

    if(!letters[letterId]){
      letters[letterId] = {
        loaded:loaded,
        letter:letter
      }
    }

    //是否存在未加载完成的字母
    let flag = false;

    for(let tKey in letters){
      if(!letters[tKey].loaded){
        flag = true;
        break;
      }
    }

    if(!flag){
      this.wordLoad[id].loaded = true;
    }
    else{
      this.wordLoad[id].loaded = false;
    }


  }

  getWordLoaded(id: string):boolean{

    if(!this.wordLoad[id]){
      return false;
    }
    return this.wordLoad[id].loaded;
  }

  /**
   * 检查单词是否加载完成
   * @param id 单词id
   */
  checkWordLoaded(id: string):void{
    if(!this.wordLoad[id]){
      return;
    }

    let letters = this.wordLoad[id].letters;

    //是否存在未加载完成的字母
    let flag = false;

    for(let tKey in letters){
      if(!letters[tKey].loaded){
        flag = true;
        break;
      }
    }

    if(!flag){
      this.wordLoad[id].loaded = true;
    }
    else{
      this.wordLoad[id].loaded = false;
    }
  }

  getLetterSpriteFromWordLoaded(id: string){

    if(!this.wordLoad[id]){
      return;
    }

    let letters = this.wordLoad[id].letters;

    return letters;
  }

  @HostListener('window:resize', ['$event'])
  onResize(event) {
    this.winResizeEventStream.next();
  }


  ngOnInit() {

    this.data = {};

    // 获取数据
    const getData = (<any> window).courseware.getData;
    getData((data) => {

      if (data && typeof data == 'object') {
        this.data = data;
      }
      // console.log('data:' , data);

      // 初始化 各事件监听
      this.initListener();

      // 若无数据 则为预览模式 需要填充一些默认数据用来显示
      this.initDefaultData();

      // 初始化 音频资源
      this.initAudio();
      // 初始化 图片资源
      this.initImg();
      // 开始预加载资源
      this.load();

      // 加载完成播放音乐
      this.playAudio("newpage");

    }, this.saveKey);

  }

  ngOnDestroy() {
    window['curCtx'] = null;
    window.cancelAnimationFrame(this.animationId);
  }


  load() {

    // 预加载资源
    this.loadResources().then(() => {
      window["air"].hideAirClassLoading(this.saveKey, this.data);
      this.init();
      this.update();
    });
  }


  init() {

    this.initCtx();
    this.initData();
    this.initView();
  }

  initCtx() {
    this.canvasWidth = this.wrap.nativeElement.clientWidth;
    this.canvasHeight = this.wrap.nativeElement.clientHeight;
    this.canvas.nativeElement.width = this.wrap.nativeElement.clientWidth;
    this.canvas.nativeElement.height = this.wrap.nativeElement.clientHeight;


    this.ctx = this.canvas.nativeElement.getContext('2d');
    this.canvas.nativeElement.width = this.canvasWidth;
    this.canvas.nativeElement.height = this.canvasHeight;

    window['curCtx'] = this.ctx;
  }






  updateItem(item) {
    if (item) {
      item.update();
    }
  }

  updateArr(arr) {
    if (!arr) {
      return;
    }
    for (let i = 0; i < arr.length; i++) {
      arr[i].update(this);
    }
  }







  initListener() {

    this.winResizeEventStream
      .pipe(debounceTime(500))
      .subscribe(data => {
        this.renderAfterResize();
      });


    // ---------------------------------------------
    const setParentOffset = () => {
      const rect = this.canvas.nativeElement.getBoundingClientRect();
      this.canvasLeft = rect.left;
      this.canvasTop = rect.top;
    };
    const setMxMyByTouch = (event) => {
      if (event.touches.length <= 0) {
        return;
      }
      if (this.canvasLeft == null) {
        setParentOffset();
      }
      this.mx = event.touches[0].pageX - this.canvasLeft;
      this.my = event.touches[0].pageY - this.canvasTop;
    };

    const setMxMyByMouse = (event) => {
      this.mx = event.offsetX;
      this.my = event.offsetY;
    };
    // ---------------------------------------------


    let firstTouch = true;

    const touchDownFunc = (e) => {
      if (firstTouch) {
        firstTouch = false;
        removeMouseListener();
      }
      setMxMyByTouch(e);
      this.mapDown(e);
    };
    const touchMoveFunc = (e) => {
      setMxMyByTouch(e);
      this.mapMove(e);
    };
    const touchUpFunc = (e) => {
      setMxMyByTouch(e);
      this.mapUp(e);
    };

    const mouseDownFunc = (e) => {
      if (firstTouch) {
        firstTouch = false;
        removeTouchListener();
      }
      setMxMyByMouse(e);
      this.mapDown(e);
    };
    const mouseMoveFunc = (e) => {
      setMxMyByMouse(e);
      this.mapMove(e);
    };
    const mouseUpFunc = (e) => {
      setMxMyByMouse(e);
      this.mapUp(e);
    };


    const element = this.canvas.nativeElement;

    const addTouchListener = () => {
      element.addEventListener('touchstart', touchDownFunc);
      element.addEventListener('touchmove', touchMoveFunc);
      element.addEventListener('touchend', touchUpFunc);
      element.addEventListener('touchcancel', touchUpFunc);
    };
    const removeTouchListener = () => {
      element.removeEventListener('touchstart', touchDownFunc);
      element.removeEventListener('touchmove', touchMoveFunc);
      element.removeEventListener('touchend', touchUpFunc);
      element.removeEventListener('touchcancel', touchUpFunc);
    };

    const addMouseListener = () => {
      element.addEventListener('mousedown', mouseDownFunc);
      element.addEventListener('mousemove', mouseMoveFunc);
      element.addEventListener('mouseup', mouseUpFunc);
    };
    const removeMouseListener = () => {
      element.removeEventListener('mousedown', mouseDownFunc);
      element.removeEventListener('mousemove', mouseMoveFunc);
      element.removeEventListener('mouseup', mouseUpFunc);
    };

    addMouseListener();
    addTouchListener();
  }


  playAudio(key, now = false, callback = null) {

    const audio = this.audioObj[key];
    if (audio) {
      if (now) {
        audio.pause();
        audio.currentTime = 0;
      }

      if (callback) {
        audio.onended = () => {
          callback();
        };
      }
      if(this.curAudio) {
        this.curAudio.pause() ;
      }
      audio.play();
      this.play = true ;
      this.curAudio = audio ;
    }
  }



  loadResources() {
    const pr = [];
    this.rawImages.forEach((value, key) => {// 预加载图片

      const p = this.preload(value)
        .then(img => {
          this.images.set(key, img);
        })
        .catch(err => console.log(err));

      pr.push(p);
    });

    this.rawAudios.forEach((value, key) => {// 预加载音频

      const a = this.preloadAudio(value)
        .then(() => {
          // this.images.set(key, img);
        })
        .catch(err => console.log(err));

      pr.push(a);
    });
    return Promise.all(pr);
  }

  preload(url) {
    return new Promise((resolve, reject) => {
      const img = new Image();
      // img.crossOrigin = "anonymous";
      img.onload = () => resolve(img);
      img.onerror = reject;
      img.src = url;
    });
  }

  preloadAudio(url) {
    return new Promise((resolve, reject) => {
      const audio = new Audio();
      audio.oncanplay = (a) => {
        resolve();
      };
      audio.onerror = () => {
        reject();
      };
      audio.src = url;
      audio.load();
    });
  }


  renderAfterResize() {
    this.canvasWidth = this.wrap.nativeElement.clientWidth;
    this.canvasHeight = this.wrap.nativeElement.clientHeight;
    this.init();
  }





  checkClickTarget(target) {

    const rect = target.getBoundingBox();

    if (this.checkPointInRect(this.mx, this.my, rect)) {
      return true;
    }
    return false;
  }

  getWorlRect(target) {

    let rect = target.getBoundingBox();

    if (target.parent) {

      const pRect = this.getWorlRect(target.parent);
      rect.x += pRect.x;
      rect.y += pRect.y;
    }
    return rect;
  }

  checkPointInRect(x, y, rect) {
    if (x >= rect.x && x <= rect.x + rect.width) {
      if (y >= rect.y && y <= rect.y + rect.height) {
        return true;
      }
    }
    return false;
  }





  addUrlToAudioObj(key, url = null, vlomue = 1, loop = false, callback = null) {

    const audioObj = this.audioObj;

    if (url == null) {
      url = key;
    }

    this.rawAudios.set(key, url);

    const audio = new Audio();
    audio.src = url;
    audio.load();
    audio.loop = loop;
    audio.volume = vlomue;

    audioObj[key] = audio;

    if (callback) {
      audio.onended = () => {
        callback();
      };
    }
  }

  addUrlToImages(url) {
    this.rawImages.set(url, url);
  }






  // ======================================================编写区域==========================================================================





  /**
   * 添加默认数据 便于无数据时的展示
   */
  initDefaultData() {

    if (!this.data.pic_url) {
      this.data.pic_url = 'assets/play/default/pic.jpg';
      this.data.pic_url_2 = 'assets/play/default/pic.jpg';
    }
  }


  /**
   * 添加预加载图片
   */
  initImg() {

    this.addUrlToImages(this.data.pic_url);
    this.addUrlToImages(this.data.pic_url_2);

  }

  /**
   * 添加预加载音频
   */
  initAudio() {
    // 音效
    this.addUrlToAudioObj('click', this.rawAudios.get('click'), 0.3);

    this.addUrlToAudioObj('newpage', this.rawAudios.get('newpage'));

    // 音频资源
    this.addUrlToAudioObj(this.data.title_audio_url);
    this.addUrlToAudioObj(this.data.words_audio_url);

    // 初始化单词音效
    for (let i = 0; i < this.data.wordArr.length; ++ i) {
       let wordLeft = this.data.wordArr[i].left;
       if (wordLeft.word_audio_url && wordLeft.word_audio_url != '') {
          this.addUrlToAudioObj(wordLeft.word_audio_url);
       }

       let wordRight = this.data.wordArr[i].right;
       if (wordRight.word_audio_url && wordRight.word_audio_url != '') {
          this.addUrlToAudioObj(wordRight.word_audio_url);
       }
    }
  }



  /**
   * 初始化数据
   */
  initData() {

    const sx = this.canvasWidth / this.canvasBaseW;
    const sy = this.canvasHeight / this.canvasBaseH;
    const s = Math.min(sx, sy);
    this.mapScale = s;

    // this.mapScale = sx;
    // this.mapScale = sy;


    this.renderArr = [];
    this.word_arr_left = [] ;
    this.word_arr = [];
    this.word_bg_x = 200 ;

    this.word_height = 0 ;

    this.all_words_loaded = false;

    this.wordLoad = {};

    this.play = false ;

  }



  /**
   * 初始化试图
   */
  initView() {


    this.initPic();

    this.initBottomPart();

  }

  initBottomPart() {

    const pause = new MySprite();
    pause.init(this.images.get('pause'));
    pause.x = this.canvasWidth - 50 * this.mapScale;
    pause.y = this.canvasHeight - 100 * this.mapScale;

    pause.setScaleXY(this.mapScale);

    this.renderArr.push(pause);

    this.pause = pause;



    // const btnRight = new MySprite();
    // btnRight.init(this.images.get('btn_right'));
    // btnRight.x = this.canvasWidth - 50 * this.mapScale;
    // btnRight.y = this.canvasHeight - 100 * this.mapScale;
    // btnRight.setScaleXY(this.mapScale);
    //
    // this.renderArr.push(btnRight);
    //
    // this.btnRight = btnRight;
  }

  initPic() {

   // 初始化单词显示
    this.initWord();

    const body_bg = new ShapeRectNew();
    let wordNum = this.data.wordArr.length;
    let line = wordNum % 2 == 0 ? wordNum / 2 :Math.ceil(wordNum / 2) ;
    // console.log('line: ', line) ;
    let beginX = 0 ;
    let beginY = 0;
    let endX = 0 ;
    let endY = 0 ;
    for (let i = 0; i < wordNum; ++ i) {
        if ( i == 0) {
           beginX = this.word_arr_left[i].x;
           beginY = this.word_arr_left[i].y;
        } else if( i == line - 1){
           endY = this.word_arr[i].y + this.word_arr[i].height;
        } else if( i == line){
           endX = this.word_arr[i].x + this.word_arr[i].width ;
        }
      body_bg.addChild(this.word_arr_left[i]) ;
      body_bg.addChild(this.word_arr[i]) ;
    }
    body_bg.setSize(1200,600,60) ;
    body_bg.x = (this.canvasWidth - body_bg.width) / 2 ;
    body_bg.y = (this.canvasHeight - body_bg.height) /2 ;
    body_bg.setOutLine('yellow',2) ;
    this.renderArr.push(body_bg);

    const words_bg = new ShapeCircle();
    words_bg.setRadius(50);
    words_bg.fillColor = '#975833' ;
    words_bg.x = this.canvasWidth / 2  ;
    words_bg.y = (this.canvasHeight - body_bg.height) /2 ;

    const words_content = new Label();
    words_content.text = this.data.words.trim().split('').join('_');
    words_content.textAlign = 'center';
    words_content.fontSize = 50;
    words_content.fontName = 'GOTHICB';
    words_content.fontColor = 'yellow';
    words_content.refreshSize();
    words_bg.addChild(words_content) ;
    this.words_bg = words_bg ;
    this.renderArr.push(this.words_bg);

    const title_bg = new ShapeRect();
    title_bg.fillColor = 'transparent' ;
    const title_left_bg = new ShapeRect();
    title_left_bg.fillColor = 'yellow' ;
    const title_left = new Label();
    title_left.text = this.data.title_left;
    title_left.textAlign = 'center';
    title_left.fontSize = this.data.title_font_size ? this.data.title_font_size : 50;
    title_left.fontName = 'GOTHICB';
    title_left.fontColor = this.getColor(this.data.title_color) ? this.getColor(this.data.title_color) : 'black';
    title_left.refreshSize();
    title_left.x = title_left.width / 2 ;
    title_left.y = title_left.height / 2 ;
    title_left_bg.setSize(title_left.width , title_left.height) ;
    title_left_bg.addChild(title_left) ;
    title_bg.addChild(title_left_bg) ;

    const title_right_bg = new ShapeRect();
    title_right_bg.fillColor = 'transparent' ;
    title_right_bg.x = title_left_bg.x+title_left_bg.width ;
    const title_right = new Label();
    title_right.text = this.data.title_right;
    title_right.textAlign = 'center';
    title_right.fontSize = this.data.title_font_size ? this.data.title_font_size : 50;
    title_right.fontName = 'GOTHICB';
    title_right.fontColor = this.getColor(this.data.title_color) ? this.getColor(this.data.title_color) : 'black';
    title_right.refreshSize();
    title_right.x = title_right.width / 2  ;
    title_right.y = title_right.height / 2 ;
    title_right_bg.setSize(title_right.width , title_right.height) ;
    title_right_bg.addChild(title_right) ;
    title_bg.setSize(title_left_bg.width + title_right_bg.width , title_right_bg.height) ;
    title_bg.addChild(title_right_bg) ;
    this.title_bg = title_bg ;
    this.renderArr.push(title_bg) ;
  }

  /**
   * 初始化单词显示
   */
  initWord() {
     this.drawWordSprite();
  }

  /**
   * 绘画单词
   */
  drawWordSprite() {

    const getWordSprite = (index , x , y) => {

      // 单词背景
      let word_bg = new ShapeRect();
      word_bg.fillColor = 'transparent' ;
      word_bg.x = x + 200 ;
      word_bg.y = y + 200;

      let start = new Label();
      start.text = (index+1)+'. ';
      start.textAlign = 'center';
      // letter_spr.fontSize = word_content.word_font_size ? word_content.word_font_size : 110;
      start.fontSize =  50;
      start.fontName = 'MMTextBookBold';
      start.fontColor = 'black';
      //设置不显示
      start.visible = true;
      start.refreshSize();
      start.y = start.height / 2 ;
      word_bg.addChild(start) ;

      //单词
      let word_content = this.data.wordArr[index];

      let totalWidth = 0;

      let height = 0;

      let letter_spr_arr = [];

      for(let i = 0; i < word_content.left.letters.length; ++ i) {



        //字母
        let letter = word_content.left.letters[i];
        // console.log('letter:',letter) ;
        let letter_spr = new Label();
        letter_spr.text = letter.letter_val;
        letter_spr.textAlign = 'center';
        // letter_spr.fontSize = word_content.word_font_size ? word_content.word_font_size : 110;
        letter_spr.fontSize =  50;
        letter_spr.fontName = 'MMTextBookBold';
        letter_spr.fontColor = this.getColor(letter.letter_color);
        //设置不显示
        letter_spr.visible = true;
        letter_spr.refreshSize();

        letter_spr_arr.push(letter_spr);

        //设置y坐标
        letter_spr.y = letter_spr.height/2;

        height = letter_spr.height;
        if(index === 0){
           this.word_height = height ;
        }
        totalWidth += letter_spr.width;

        if((i + 1) < word_content.left.letters.length){
          totalWidth += 0;
        }

        // 设置单词未加截完
        // this.setWordLoad(index, String(i), false, letter_spr);

        word_bg.addChild(letter_spr);
      }

      let temp_startX = 0;

      //设置所有字母的x坐标
      for(let i = 0; i < letter_spr_arr.length; ++ i) {

        let startX = start.width/2;

        if(i == 0){
          startX = letter_spr_arr[i].width/2 + start.width/2 ;
        }
        else{
          temp_startX += letter_spr_arr[i-1].width + 0;
          startX = temp_startX + letter_spr_arr[i].width/2+start.width/2;
        }

        letter_spr_arr[i].x = startX;

      }

      word_bg.setSize(totalWidth, height);
      // 设置不显示
      // word_bg.alpha = 0;
      // this.word_height = word_bg.height;
      this.word_arr_left.push(word_bg);
      // this.renderArr.push(word_bg) ;
      // console.log('word_arr', this.word_arr) ;

      // 单词背景
      let word_bg_right = new ShapeRect();
      word_bg_right.fillColor = 'transparent' ;
      word_bg_right.x =  word_bg.x+200;
      word_bg_right.y = y + 200;

      //单词

      let totalWidth_right = 0;

      let height_right = 0;

      let letter_spr_arr_right = [];

      for(let i = 0; i < word_content.right.letters.length; ++ i) {



        //字母
        let letter = word_content.right.letters[i];
        // console.log('letter:',letter) ;
        let letter_spr = new Label();
        letter_spr.text = letter.letter_val;
        letter_spr.textAlign = 'center';
        // letter_spr.fontSize = word_content.word_font_size ? word_content.word_font_size : 110;
        letter_spr.fontSize =  50;
        letter_spr.fontName = 'MMTextBookBold';
        letter_spr.fontColor = this.getColor(letter.letter_color);
        //设置不显示
        letter_spr.visible = true;
        letter_spr.refreshSize();

        letter_spr_arr_right.push(letter_spr);

        //设置y坐标
        letter_spr.y = letter_spr.height/2;

        height_right = letter_spr.height;
        totalWidth_right += letter_spr.width;

        if((i + 1) < word_content.right.letters.length){
          totalWidth_right += 0;
        }

        // 设置单词未加截完
        this.setWordLoad(index, String(i), false, letter_spr);

        word_bg_right.addChild(letter_spr);
      }

      let temp_startX_right = 0;

      //设置所有字母的x坐标
      for(let i = 0; i < letter_spr_arr_right.length; ++ i) {

        let startX = letter_spr_arr[letter_spr_arr.length - 1].x;

        if(i == 0){
          startX = letter_spr_arr[i].width/2;
        }
        else{
          temp_startX_right += letter_spr_arr_right[i-1].width + 0;
          startX = temp_startX_right + letter_spr_arr_right[i].width/2;
        }

        letter_spr_arr_right[i].x = startX;

        if ( letter_spr_arr_right[i].text === this.data.words.trim().split('')[0] || letter_spr_arr_right[i].text === this.data.words.trim().split('')[1]) {
          letter_spr_arr_right[i].visible = false ;
          let label = new Label();
          label.text = '_';
          label.textAlign = 'center';
          // letter_spr.fontSize = word_content.word_font_size ? word_content.word_font_size : 110;
          label.fontSize =  50;
          label.fontName = 'MMTextBookBold';
          label.fontColor = 'red';
          //设置不显示
          label.visible = true;
          label.refreshSize();
          label.x = letter_spr_arr_right[i].x ;
          label.y = letter_spr_arr_right[i].height / 2 ;
          word_bg_right.addChild(label) ;
        }

      }

      word_bg_right.setSize(totalWidth_right, height_right);
      // 设置不显示
      // word_bg.alpha = 0;
      // this.word_height = word_bg.height;
      this.word_arr.push(word_bg_right);
      // this.renderArr.push(word_bg_right) ;
    };

    //单词个数
    let wordNum = this.data.wordArr.length;
    let line = wordNum % 2 == 0 ? wordNum / 2 :Math.ceil(wordNum / 2) ;
    // console.log('line: ', line) ;
    for (let i = 0; i < wordNum; ++ i) {
        if(i <= line-1){
          getWordSprite(i , 0 , this.word_height*i*1.3);
        } else {
          getWordSprite(i , 500 , this.word_height*(i-line)*1.3);
        }

    }
  }

  /**
   * 根据颜色id获取对应的颜色
   * @param colorId 颜色id
   */
  getColor(colorId){

    let color = "#000000";

    switch(colorId){
      case "C01":
        color = "#008000";
        break;
      case "C02":
        color = "#FF0000";
        break;
      case "C03":
        color = "#D2691E";
        break;
      case "C04":
        color = "#000000";
        break;
    }

    return color;
  }

  pauseClicked() {
     if(this.curAudio){
         if(this.play){
            this.curAudio.pause() ;
            this.play = false ;
         }else {
           this.curAudio.play() ;
           this.play = true ;
         }
     }
  }

  btnRightClicked() {

    this.nextPage();
  }

  lastPage() {

    if (this.curPic == this.pic1) {
      return;
    }

    this.canTouch = false;

    const moveLen = this.canvasWidth;
    tweenChange(this.pic1, {x: this.pic1.x + moveLen}, 1);
    tweenChange(this.pic2, {x: this.pic2.x + moveLen}, 1, () => {
      this.canTouch = true;
      this.curPic = this.pic1;
    });
  }

  nextPage() {

    if (this.curPic == this.pic2) {
      return;
    }

    this.canTouch = false;

    const moveLen = this.canvasWidth;
    tweenChange(this.pic1, {x: this.pic1.x - moveLen}, 1);
    tweenChange(this.pic2, {x: this.pic2.x - moveLen}, 1, () => {
      this.canTouch = true;
      this.curPic = this.pic2;
    });
  }

  pic1Clicked() {
    this.playAudio(this.data.audio_url);
  }

  pic2Clicked() {
    this.playAudio(this.data.audio_url_2);
  }





  mapDown(event) {

    if (!this.canTouch) {
      return;
    }

    if (this.checkClickTarget(this.title_bg)) {
      // console.log('click the words :' , this.data.words_audio_url) ;
      this.playAudio(this.data.title_audio_url) ;
      return ;
    }

    if (this.checkClickTarget(this.words_bg)) {
        // console.log('click the words :' , this.data.words_audio_url) ;
        this.playAudio(this.data.words_audio_url) ;
        return ;
    }

    for(let i = 0; i < this.word_arr_left.length; ++ i){
      let word = this.word_arr_left[i];
      // console.log('word', word) ;
      if (this.checkClickTarget(word)) {
        console.log('this word left: ', this.data.wordArr[i]) ;
        this.playAudio(this.data.wordArr[i].left.word_audio_url);
        return ;
      }

    }

    for(let i = 0; i < this.word_arr.length; ++ i){
      let word = this.word_arr[i];
      // console.log('word', word) ;
      if (this.checkClickTarget(word)) {
        // 获取是否加载完成
        let loaded = this.getWordLoaded(String(i));
        // 未加载，不触发音效，加载单词
        if ( !loaded) {
             let letters = this.getLetterSpriteFromWordLoaded(String(i));
             for (let key in letters) {
               let loaded = letters[key].loaded;
               let letter_spr = letters[key].letter;
               //加载完成，跳过
               if(loaded){
                 continue;
               }
               //设置显示
               letter_spr.visible = true;
               //设置加载完成
               letters[key].loaded = true;
               this.checkWordLoaded(String(i));
             }
             return ;
        }
        console.log('this word: ', this.data.wordArr[i]) ;
        this.playAudio(this.data.wordArr[i].right.word_audio_url);
        return ;
      }

    }

    if ( this.checkClickTarget(this.pause) ) {
      this.pauseClicked();
      return;
    }

    if ( this.checkClickTarget(this.btnRight) ) {
      this.btnRightClicked();
      return;
    }

    if ( this.checkClickTarget(this.pic1) ) {
      this.pic1Clicked();
      return;
    }

    if ( this.checkClickTarget(this.pic2) ) {
      this.pic2Clicked();
      return;
    }

  }

  mapMove(event) {

  }

  mapUp(event) {

  }



  update() {

    // ----------------------------------------------------------
    this.animationId = window.requestAnimationFrame(this.update.bind(this));
    // 清除画布内容
    this.ctx.clearRect(0, 0, this.canvasWidth, this.canvasHeight);
    // tween 更新动画
    TWEEN.update();
    // ----------------------------------------------------------



    this.updateArr(this.renderArr);


  }



}
